#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <complex.h>
#include "matrix.h"
#include "exponentialsum.h"
#include "shrink.h"
#include "extend.h"
#include "measurepauli.h"
#include "innerproduct.h"

int readPaulicoeffs(int *omega, int *alpha, int *beta, int *gamma, int *delta, int numqubits);

// order of matrix elements is [row][column]!!!

int main()
{

  int N;              // number of qubits
  scanf("%d", &N);

  if(N%6 != 0) {
    printf("'N' needs to be a multiple of 6 for a k=6 tensor factor decomposition!\n");
    return 1;
  }

  int T;              // number of T gate magic states (set to the first 'K' of the 'N' qubits -- the rest are set to the '0' computational basis state)
  scanf("%d", &T);  

  int omega;
  int alpha[N], beta[N], gamma[N], delta[N];
  int Paulicounter = 0;

  int i, j, k, l, m;

  double complex coeffb60 = (-16.0+12.0*sqrt(2.0)+0.0*I)*pow(cos(PI/8.0),6)*cexp(6.0*PI*I/8.0)/8.0*cpow(2.0,3.0);
  double complex coeffb66 = (96.0-68.0*sqrt(2.0)+0.0*I)*pow(cos(PI/8.0),6)*cexp(6.0*PI*I/8.0)/8.0*cpow(2.0,3.0);
  double complex coeffe6 = (10.0-7.0*sqrt(2.0)+0.0*I)*pow(cos(PI/8.0),6)*cexp(6.0*PI*I/8.0)*cpow(2.0,2.5);
  double complex coeffo6 = (-14.0+10.0*sqrt(2.0)+0.0*I)*pow(cos(PI/8.0),6)*cexp(-14.0*PI*I/8.0)*cpow(2.0,2.5);
  double complex coeffk6 = (7.0-5.0*sqrt(2.0)+0.0*I)*pow(cos(PI/8.0),6)*cexp(-8.0*PI*I/8.0)*4.0*csqrt(2.0)*cpow(2.0,0.5);
  double complex coeffphiprime = (10.0-7.0*sqrt(2.0)+0.0*I)*pow(cos(PI/8.0),6)*cexp(2.0*PI*I/8.0)*cpow(2.0,2.5);
  double complex coeffphidprime = (10.0-7.0*sqrt(2.0)+0.0*I)*pow(cos(PI/8.0),6)*cexp(2.0*PI*I/8.0)*cpow(2.0,2.5);

  // b60
  int n1 = 6; int k1 = 6; int (*(G1[])) = { (int[]) {1, 0, 0, 0, 0, 0}, (int[]) {0, 1, 0, 0, 0, 0}, (int[]) {0, 0, 1, 0, 0, 0}, (int[]) {0, 0, 0, 1, 0, 0}, (int[]) {0, 0, 0, 0, 1, 0}, (int[]) {0, 0, 0, 0, 0, 1}}; int (*(GBar1[])) = { (int[]) {1, 0, 0, 0, 0, 0}, (int[]) {0, 1, 0, 0, 0, 0}, (int[]) {0, 0, 1, 0, 0, 0}, (int[]) {0, 0, 0, 1, 0, 0}, (int[]) {0, 0, 0, 0, 1, 0}, (int[]) {0, 0, 0, 0, 0, 1}}; int h1[] = {0, 0, 0, 0, 0, 0}; int Q1 = 0; int D1[] = {0, 0, 0, 0, 0, 0}; int (*(J1[])) = { (int[]) {0, 0, 0, 0, 0, 0}, (int[]) {0, 0, 0, 0, 0, 0}, (int[]) {0, 0, 0, 0, 0, 0}, (int[]) {0, 0, 0, 0, 0, 0}, (int[]) {0, 0, 0, 0, 0, 0}, (int[]) {0, 0, 0, 0, 0, 0} };
  // b66
  int n2 = 6; int k2 = 6; int (*(G2[])) = { (int[]) {1, 0, 0, 0, 0, 0}, (int[]) {0, 1, 0, 0, 0, 0}, (int[]) {0, 0, 1, 0, 0, 0}, (int[]) {0, 0, 0, 1, 0, 0}, (int[]) {0, 0, 0, 0, 1, 0}, (int[]) {0, 0, 0, 0, 0, 1}}; int (*(GBar2[])) = { (int[]) {1, 0, 0, 0, 0, 0}, (int[]) {0, 1, 0, 0, 0, 0}, (int[]) {0, 0, 1, 0, 0, 0}, (int[]) {0, 0, 0, 1, 0, 0}, (int[]) {0, 0, 0, 0, 1, 0}, (int[]) {0, 0, 0, 0, 0, 1}}; int h2[] = {0, 0, 0, 0, 0, 0}; int Q2 = 4; int D2[] = {4, 4, 4, 4, 4, 4}; int (*(J2[])) = { (int[]) {0, 0, 0, 0, 0, 0}, (int[]) {0, 0, 0, 0, 0, 0}, (int[]) {0, 0, 0, 0, 0, 0}, (int[]) {0, 0, 0, 0, 0, 0}, (int[]) {0, 0, 0, 0, 0, 0}, (int[]) {0, 0, 0, 0, 0, 0} };
  // e6
  int n3 = 6; int k3 = 5; int (*(G3[])) = { (int[]) {1, 1, 0, 0, 0, 0}, (int[]) {1, 0, 1, 0, 0, 0}, (int[]) {1, 0, 0, 1, 0, 0}, (int[]) {1, 0, 0, 0, 1, 0}, (int[]) {1, 0, 0, 0, 0, 1}, (int[]) {1, 0, 0, 0, 0, 0} }; int (*(GBar3[])) = { (int[]) {0, 1, 0, 0, 0, 0}, (int[]) {0, 0, 1, 0, 0, 0}, (int[]) {0, 0, 0, 1, 0, 0}, (int[]) {0, 0, 0, 0, 1, 0}, (int[]) {0, 0, 0, 0, 0, 1}, (int[]) {1, 1, 1, 1, 1, 1} }; int h3[] = {1, 0, 0, 0, 0, 0}; int Q3 = 4; int D3[] = {0, 0, 0, 0, 0}; int (*(J3[])) = { (int[]) {0, 4, 4, 4, 4}, (int[]) {4, 0, 4, 4, 4}, (int[]) {4, 4, 0, 4, 4}, (int[]) {4, 4, 4, 0, 4}, (int[]) {4, 4, 4, 4, 0}  };
  // o6
  int n4 = 6; int k4 = 5; int (*(G4[])) = { (int[]) {1, 1, 0, 0, 0, 0}, (int[]) {1, 0, 1, 0, 0, 0}, (int[]) {1, 0, 0, 1, 0, 0}, (int[]) {1, 0, 0, 0, 1, 0}, (int[]) {1, 0, 0, 0, 0, 1}, (int[]) {1, 0, 0, 0, 0, 0} }; int (*(GBar4[])) = { (int[]) {0, 1, 0, 0, 0, 0}, (int[]) {0, 0, 1, 0, 0, 0}, (int[]) {0, 0, 0, 1, 0, 0}, (int[]) {0, 0, 0, 0, 1, 0}, (int[]) {0, 0, 0, 0, 0, 1}, (int[]) {1, 1, 1, 1, 1, 1} }; int h4[] = {0, 0, 0, 0, 0, 0}; int Q4 = 4; int D4[] = {4, 4, 4, 4, 4}; int (*(J4[])) = { (int[]) {0, 4, 4, 4, 4}, (int[]) {4, 0, 4, 4, 4}, (int[]) {4, 4, 0, 4, 4}, (int[]) {4, 4, 4, 0, 4}, (int[]) {4, 4, 4, 4, 0}  };
  // k6
  int n5 = 6; int k5 = 1; int (*(G5[])) = { (int[]) {1, 1, 1, 1, 1, 1}, (int[]) {0, 1, 0, 0, 0, 0}, (int[]) {0, 0, 1, 0, 0, 0}, (int[]) {0, 0, 0, 1, 0, 0}, (int[]) {0, 0, 0, 0, 1, 0}, (int[]) {0, 0, 0, 0, 0, 1} }; int (*(GBar5[])) = { (int[]) {1, 0, 0, 0, 0, 0}, (int[]) {1, 1, 0, 0, 0, 0}, (int[]) {1, 0, 1, 0, 0, 0}, (int[]) {1, 0, 0, 1, 0, 0}, (int[]) {1, 0, 0, 0, 1, 0}, (int[]) {1, 0, 0, 0, 0, 1} }; int h5[] = {1, 1, 1, 1, 1, 1}; int Q5 = 6; int D5[] = {2}; int (*(J5[])) = { (int[]) {4} };
  // phiprime
  int n6 = 6; int k6 = 5; int (*(G6[])) = { (int[]) {1, 1, 0, 0, 0, 0}, (int[]) {1, 0, 1, 0, 0, 0}, (int[]) {1, 0, 0, 1, 0, 0}, (int[]) {1, 0, 0, 0, 1, 0}, (int[]) {1, 0, 0, 0, 0, 1}, (int[]) {1, 0, 0, 0, 0, 0} }; int (*(GBar6[])) = { (int[]) {0, 1, 0, 0, 0, 0}, (int[]) {0, 0, 1, 0, 0, 0}, (int[]) {0, 0, 0, 1, 0, 0}, (int[]) {0, 0, 0, 0, 1, 0}, (int[]) {0, 0, 0, 0, 0, 1}, (int[]) {1, 1, 1, 1, 1, 1} }; int h6[] = {1, 0, 0, 0, 0, 0}; int Q6 = 0; int D6[] = {0, 0, 0, 0, 0}; int (*(J6[])) = { (int[]) {0, 4, 0, 0, 4}, (int[]) {4, 0, 4, 0, 0}, (int[]) {0, 4, 0, 4, 0}, (int[]) {0, 0, 4, 0, 4}, (int[]) {4, 0, 0, 4, 0}  };
  // phidoubleprime
  int n7 = 6; int k7 = 5; int (*(G7[])) = { (int[]) {1, 1, 0, 0, 0, 0}, (int[]) {1, 0, 1, 0, 0, 0}, (int[]) {1, 0, 0, 1, 0, 0}, (int[]) {1, 0, 0, 0, 1, 0}, (int[]) {1, 0, 0, 0, 0, 1}, (int[]) {1, 0, 0, 0, 0, 0} }; int (*(GBar7[])) = { (int[]) {0, 1, 0, 0, 0, 0}, (int[]) {0, 0, 1, 0, 0, 0}, (int[]) {0, 0, 0, 1, 0, 0}, (int[]) {0, 0, 0, 0, 1, 0}, (int[]) {0, 0, 0, 0, 0, 1}, (int[]) {1, 1, 1, 1, 1, 1} }; int h7[] = {1, 0, 0, 0, 0, 0}; int Q7 = 0; int D7[] = {0, 0, 0, 0, 0}; int (*(J7[])) = { (int[]) {0, 0, 4, 4, 0}, (int[]) {0, 0, 0, 4, 4}, (int[]) {4, 0, 0, 0, 4}, (int[]) {4, 4, 0, 0, 0}, (int[]) {0, 4, 4, 0, 0}  };
  
  int *K; int ***G; int ***GBar; int **h; int *Q; int **D; int ***J;
  double complex Gamma[(int)pow(7,N/6)]; // prefactor in front of resultant state
  G = calloc(pow(7,N/6),sizeof(int*)); GBar = calloc(pow(7,N/6),sizeof(int*));
  h = calloc(pow(7,N/6),sizeof(int*));
  
  J = calloc(pow(7,N/6),sizeof(int*)); D = calloc(pow(7,N/6),sizeof(int*)); Q = calloc(pow(7,N/6),sizeof(int));

  K = calloc(pow(7,N/6), sizeof(int));

  double complex origGamma[(int)pow(7,N/6)];
  int *origK, *origQ, **origD, ***origJ;
  int ***origG, ***origGBar, **origh;

  origG = calloc(pow(7,N/6),sizeof(int*)); origGBar = calloc(pow(7,N/6),sizeof(int*));
  origh = calloc(pow(7,N/6),sizeof(int*));
  
  origJ = calloc(pow(7,N/6),sizeof(int*)); origD = calloc(pow(7,N/6),sizeof(int*)); origQ = calloc(pow(7,N/6),sizeof(int));

  origK = calloc(pow(7,N/6), sizeof(int));

  int combination; // a particular combination from the linear combo of stabilizer states making up the tensor factors multiplied together
  

  for(j=0; j<pow(7,N/6); j++) { // there will be 7^(N/6) combinations when using k=3 tensor factors

    combination = j;

    K[j] = 0.0;
    
    for(k=0; k<N/6; k++) {
      K[j] += (((combination%7)==6)*k7 + ((combination%7)==5)*k6 + ((combination%7)==4)*k5 + ((combination%7)==3)*k4 + ((combination%7)==2)*k3 + ((combination%7)==1)*k2 + ((combination%7)==0)*k1);
      combination /= 7;
    }
    combination = j;
    origK[j] = K[j];

    Gamma[j] = 1.0;

    G[j] = calloc(N, sizeof(int*)); GBar[j] = calloc(N, sizeof(int*));
    h[j] = calloc(N, sizeof(int));

    if(K[j] > 0) {
      J[j] = calloc(K[j], sizeof(int*)); D[j] = calloc(K[j], sizeof(int));
      for(k=0; k<K[j]; k++)
	J[j][k] = calloc(K[j], sizeof(int));
    }

    origG[j] = calloc(N, sizeof(int*)); origGBar[j] = calloc(N, sizeof(int*));
    origh[j] = calloc(N, sizeof(int));

    if(K[j] > 0) {
      origJ[j] = calloc(K[j], sizeof(int*)); origD[j] = calloc(K[j], sizeof(int));
      for(k=0; k<K[j]; k++)
	origJ[j][k] = calloc(K[j], sizeof(int));
    }

    for(k=0; k<N; k++) {
      G[j][k] = calloc(N, sizeof(int)); GBar[j][k] = calloc(N, sizeof(int));
      origG[j][k] = calloc(N, sizeof(int)); origGBar[j][k] = calloc(N, sizeof(int));
    }

    int Kcounter = 0; // Kcounter keeps track of the K<=N that we have added already to the G rows etc. for each combination that is indexed by the digits (base 3) of 'j' in that we go through with 'k'
    int Kcombo; // Kcombo stores the k<(n1=n2=n3) dimension of the member of the combination that we are currently adding
    for(k=0; k<N/6; k++) {

      Q[j] += ((combination%7)==6)*Q7 + ((combination%7)==5)*Q6 + ((combination%7)==4)*Q5 + ((combination%7)==3)*Q4 + ((combination%7)==2)*Q3 + ((combination%7)==1)*Q2 + ((combination%7)==0)*Q1;
      
      Gamma[j] *= (((combination%7)==6)*coeffphidprime + ((combination%7)==5)*coeffphiprime + ((combination%7)==4)*coeffk6 + ((combination%7)==3)*coeffo6 + ((combination%7)==2)*coeffe6 + ((combination%7)==1)*coeffb66 + ((combination%7)==0)*coeffb60);

      Kcombo = (((combination%7)==6)*k7 + ((combination%7)==5)*k6 + ((combination%7)==4)*k5 + ((combination%7)==3)*k4 + ((combination%7)==2)*k3 + ((combination%7)==1)*k2 + ((combination%7)==0)*k1);
      for(l=0; l<Kcombo; l++) {
	// D1 has a different number of rows 'l' than D2 and D3 so you need to use something like 'switch' to check combination%3 without going out of bound of J1
	switch(combination%7) {
	case 0:
	  D[j][Kcounter+l] = D1[l];
	  break;
	case 1:
	  D[j][Kcounter+l] = D2[l];
	  break;
	case 2:
	  D[j][Kcounter+l] = D3[l];
	  break;
	case 3:
	  D[j][Kcounter+l] = D4[l];
	    break;
	case 4:
	  D[j][Kcounter+l] = D5[l];
	  break;
	case 5:
	  D[j][Kcounter+l] = D6[l];
	  break;
	case 6:
	  D[j][Kcounter+l] = D7[l];
	  break;
	default:
	  printf("error");
	  return 1;
	  }
	for(m=0; m<Kcombo; m++) {
	  // J1 has a different number of rows 'l' than J2 and J3 so you need to use something like 'switch' to check combination%3 without going out of bound of J1
	  switch(combination%7) {
	  case 0:
	    J[j][Kcounter+l][Kcounter+m] = J1[l][m];
	    break;
	  case 1:
	      J[j][Kcounter+l][Kcounter+m] = J2[l][m];
	      break;
	  case 2:
	    J[j][Kcounter+l][Kcounter+m] = J3[l][m];
	    break;
	  case 3:
	    J[j][Kcounter+l][Kcounter+m] = J4[l][m];
	    break;
	    case 4:
	      J[j][Kcounter+l][Kcounter+m] = J5[l][m];
	      break;
	  case 5:
	    J[j][Kcounter+l][Kcounter+m] = J6[l][m];
	    break;
	  case 6:
	    J[j][Kcounter+l][Kcounter+m] = J7[l][m];
	    break;
	  default:
	    printf("error");
	    return 1;
	  }
	}
      }

      for(l=0; l<n1; l++) { // assuming n1=n2=n3
	h[j][k*n1+l] = ((combination%7)==6)*h7[l] + ((combination%7)==5)*h6[l] + ((combination%7)==4)*h5[l] + ((combination%7)==3)*h4[l] + ((combination%7)==2)*h3[l] + ((combination%7)==1)*h2[l] + ((combination%7)==0)*h1[l];
      }
      // only filling the K[j] first rows of G and GBar here corresponding to the basis for D and J
      for(l=0; l<Kcombo; l++) {
	for(m=0; m<n1; m++) { // assuming n1=n2=n3
	  G[j][Kcounter+l][k*n1+m] = ((combination%7)==6)*G7[l][m] + ((combination%7)==5)*G6[l][m] + ((combination%7)==4)*G5[l][m] + ((combination%7)==3)*G4[l][m] + ((combination%7)==2)*G3[l][m] + ((combination%7)==1)*G2[l][m] + ((combination%7)==0)*G1[l][m];
	  GBar[j][Kcounter+l][k*n1+m] = ((combination%7)==6)*GBar7[l][m] + ((combination%7)==5)*GBar6[l][m] + ((combination%7)==4)*GBar5[l][m] + ((combination%7)==3)*GBar4[l][m] + ((combination%7)==2)*GBar3[l][m] + ((combination%7)==1)*GBar2[l][m] + ((combination%7)==0)*GBar1[l][m];
	}
      }
      Kcounter = Kcounter + Kcombo;
      
      /* printf("intermediate G[%d]:\n", j); */
      /* printMatrix(G[j], N, N); */
      /* printf("intermediate GBar[%d]:\n", j); */
      /* printMatrix(GBar[j], N, N); */
      //memcpy(origG[j][k], G[j][k], N*sizeof(int)); memcpy(origGBar[j][k], GBar[j][k], N*sizeof(int));

      //memcpy(origJ[j][k], J[j][k], K[j]*sizeof(int));
      
      combination /= 7; // shift to the right by one (in base-7 arithmetic)
    }
    //printf("!\n");

    // now need to fill the N-Kcounter remaining rows of G and GBar that are outside the spanning basis states of D and J
    combination = j;
    for(k=0; k<(N/6); k++) {
      Kcombo = (((combination%7)==6)*k7 + ((combination%7)==5)*k6 + ((combination%7)==4)*k5 + ((combination%7)==3)*k4 + ((combination%7)==2)*k3 + ((combination%7)==1)*k2 + ((combination%7)==0)*k1);
      //printf("Kcounter=%d\n", Kcounter);
      // G and GBar rows that are outside the first 'k' spanning basis states
      for(l=Kcombo; l<n1; l++) { // assuming n1=n2=n3
	//printf("l=%d\n", l);
      	for(m=0; m<n1; m++) { // assuming n1=n2=n3
	  /* printf("m=%d\n", m); */
	  /* printf("Kcounter+l=%d\n", Kcounter+l); */
	  /* printf("k*n1+m=%d\n", k*n1+m); */
      	  G[j][Kcounter+l-Kcombo][k*n1+m] = ((combination%7)==6)*G7[l][m] + ((combination%7)==5)*G6[l][m] + ((combination%7)==4)*G5[l][m] + ((combination%7)==3)*G4[l][m] + ((combination%7)==2)*G3[l][m] + ((combination%7)==1)*G2[l][m] + ((combination%7)==0)*G1[l][m];
      	  GBar[j][Kcounter+l-Kcombo][k*n1+m] = ((combination%7)==6)*GBar7[l][m] + ((combination%7)==5)*GBar6[l][m] + ((combination%7)==4)*GBar5[l][m] + ((combination%7)==3)*GBar4[l][m] + ((combination%7)==2)*GBar3[l][m] + ((combination%7)==1)*GBar2[l][m] + ((combination%7)==0)*GBar1[l][m];
      	}
      }
      Kcounter = Kcounter + (n1-Kcombo);

      /* printf("intermediate G[%d]:\n", j); */
      /* printMatrix(G[j], N, N); */
      /* printf("intermediate GBar[%d]:\n", j); */
      /* printMatrix(GBar[j], N, N); */
      
      combination /= 7;
    }
    for(k=0; k<N; k++) {
      memcpy(origG[j][k], G[j][k], N*sizeof(int)); memcpy(origGBar[j][k], GBar[j][k], N*sizeof(int));
    }
    for(k=0; k<K[j]; k++) {
      memcpy(origJ[j][k], J[j][k], K[j]*sizeof(int));      
    }

    memcpy(origh[j], h[j], N*sizeof(int));
    memcpy(origD[j], D[j], K[j]*sizeof(int));

    printf("G[%d]:\n", j);
    printMatrix(G[j], N, N);
    printf("GBar[%d]:\n", j);
    printMatrix(GBar[j], N, N);

    printf("h[%d]:\n", j);
    printVector(h[j], N);

    printf("J[%d]:\n", j);
    printMatrix(J[j], K[j], K[j]);
    
    printf("D[%d]:\n", j);
    printVector(D[j], K[j]);

    printf("Q[%d]=%d\n", j, Q[j]);

  }
  //exit(0);
  memcpy(origGamma, Gamma, pow(7,N/6)*sizeof(double complex));

  memcpy(origQ, Q, pow(7,N/6)*sizeof(int));

  while(readPaulicoeffs(&omega, alpha, beta, gamma, delta, N)) {
  
    Paulicounter++;
    if(Paulicounter > N) {
      printf("Error: Number of Paulis is greater than N!\n");
      return 1;
    }
    
    // Let's break up the Ys into Xs and Zs in the order Z X, as required to pass to measurepauli()
    // Y_i = -I*Z*X
    for(i=0; i<N; i++) {
      if(delta[i]){
	omega += 3; // -I = I^3
	beta[i] = delta[i];
	gamma[i] = delta[i];
      }
    }

    printf("*******\n");
    printf("*******\n");
    printf("omega=%d\n", omega);
    printf("X:\n");
    printVector(gamma, N);
    printf("Z:\n");
    printVector(beta, N);
    printf("*******\n");
    printf("*******\n");

    for(j=0; j<pow(7,N/6); j++) { // the kets

      printf("========\n");
      printf("before:\n");
      printf("K=%d\n", K[j]);
      printf("h:\n");
      printVector(h[j], N);
      printf("Gamma[%d]=%lf+%lf\n", j, creal(Gamma[j]), cimag(Gamma[j]));
      printf("G:\n");
      printMatrix(G[j], N, N);
      printf("GBar:\n");
      printMatrix(GBar[j], N, N);
      printf("Q=%d\n", Q[j]);
      printf("D:\n");
      printVector(D[j], K[j]);
      printf("J:\n");
      printMatrix(J[j], K[j], K[j]);
      Gamma[j] *= measurepauli(N, &K[j], h[j], G[j], GBar[j], &Q[j], &D[j], &J[j], omega, gamma, beta);
      printf("\nafter:\n");
      printf("K=%d\n", K[j]);
      printf("h:\n");
      printVector(h[j], N);
      printf("Gamma[%d]=%lf+%lf\n", j, creal(Gamma[j]), cimag(Gamma[j]));
      printf("G:\n");
      printMatrix(G[j], N, N);
      printf("GBar:\n");
      printMatrix(GBar[j], N, N);
      printf("Q=%d\n", Q[j]);
      printf("D:\n");
      printVector(D[j], K[j]);
      printf("J:\n");
      printMatrix(J[j], K[j], K[j]);

    }

  }

  double complex amplitude = 0.0 + 0.0*I;
  for(i=0; i<pow(7,N/6); i++) { // the bras
    for(j=0; j<pow(7,N/6); j++) {
      double complex newamplitude = innerproduct(N, K[j], h[j], G[j], GBar[j], Q[j], D[j], J[j], N, origK[i], origh[i], origG[i], origGBar[i], origQ[i], origD[i], origJ[i]);
      amplitude = amplitude + conj(origGamma[i])*Gamma[j]*newamplitude;
    }
  }

  printf("amplitude:\n");
  if(creal(amplitude+0.00000001)>0)
    printf("%.10lf %c %.10lf I\n", cabs(creal(amplitude)), cimag(amplitude+0.00000001)>0?'+':'-' , cabs(cimag(amplitude)));
  else
    printf("%.10lf %c %.10lf I\n", creal(amplitude), cimag(amplitude+0.00000001)>0?'+':'-' , cabs(cimag(amplitude)));

  return 0;

}



int readPaulicoeffs(int *omega, int *alpha, int *beta, int *gamma, int *delta, int numqubits)
{
    
  int newomega, newalpha, newbeta, newgamma, newdelta;
  int i;

  if(scanf("%d", &newomega) != EOF) {
    *omega = newomega;
    for(i=0; i<numqubits; i++) {
      if(scanf("%d %d %d %d", &newalpha, &newbeta, &newgamma, &newdelta) == EOF) {
	printf("Error: Too few input coeffs!\n");
	exit(0);
      }
      if(newalpha+newbeta+newgamma+newdelta > 1) {
	printf("Error: Too many coefficients are non-zero at Pauli %d!\n", i);
	exit(0);
      }
      alpha[i] = newalpha; beta[i] = newbeta; gamma[i] = newgamma; delta[i] = newdelta;
    }
    return 1;
  } else
    return 0;
    
}
